@@ -13,12 +13,13 @@
 ##############################################################################
 """Test runner
 
-$Id: testrunner.py 78154 2007-07-19 14:14:33Z mgedmin $
+$Id: testrunner.py 78821 2007-08-14 22:11:05Z jim $
 """
 
 # Too bad: For now, we depend on zope.testing.  This is because
 # we want to use the latest, greatest doctest, which zope.testing
 # provides.  Then again, zope.testing is generally useful.
+
 import gc
 import glob
 import logging
@@ -27,6 +28,7 @@ import os
 import errno
 import pdb
 import re
+import cStringIO
 import sys
 import tempfile
 import threading
@@ -259,6 +261,33 @@ Got:
 """
 
 
+def tigetnum(attr, default=None):
+    """Return a value from the terminfo database.
+
+    Terminfo is used on Unix-like systems to report various terminal attributes
+    (such as width, height or the number of supported colors).
+
+    Returns ``default`` when the ``curses`` module is not available, or when
+    sys.stdout is not a terminal.
+    """
+    try:
+        import curses
+    except ImportError:
+        # avoid reimporting a broken module in python 2.3
+        sys.modules['curses'] = None
+    else:
+        try:
+            curses.setupterm()
+        except (curses.error, TypeError):
+            # You get curses.error when $TERM is set to an unknown name
+            # You get TypeError when sys.stdout is not a real file object
+            # (e.g. in unit tests that use various wrappers).
+            pass
+        else:
+            return curses.tigetnum(attr)
+    return default
+
+
 class OutputFormatter(object):
     """Test runner output formatter."""
 
@@ -280,19 +309,8 @@ class OutputFormatter(object):
 
     def compute_max_width(self):
         """Try to determine the terminal width."""
-        try:
-            # Note that doing this every time is more test friendly.
-            import curses
-        except ImportError:
-            # avoid reimporting a broken module in python 2.3
-            sys.modules['curses'] = None
-        else:
-            try:
-                curses.setupterm()
-            except TypeError:
-                pass
-            else:
-                self.max_width = curses.tigetnum('cols')
+        # Note that doing this every time is more test friendly.
+        self.max_width = tigetnum('cols', self.max_width)
 
     def getShortDescription(self, test, room):
         """Return a description of a test that fits in ``room`` characters."""
@@ -601,6 +619,7 @@ class ColorfulOutputFormatter(OutputFormatter):
                    'suboptimal-behaviour': 'magenta',
                    'error': 'brightred',
                    'number': 'green',
+                   'slow-test': 'brightmagenta',
                    'ok-number': 'green',
                    'error-number': 'brightred',
                    'filename': 'lightblue',
@@ -642,6 +661,8 @@ class ColorfulOutputFormatter(OutputFormatter):
                   'cyan': 36,
                   'grey': 37, 'gray': 37, 'white': 37}
 
+    slow_test_threshold = 10.0 # seconds
+
     def color_code(self, color):
         """Convert a color description (e.g. 'lightgray') to a terminal code."""
         prefix_code = ''
@@ -710,6 +731,14 @@ class ColorfulOutputFormatter(OutputFormatter):
             return "%s seconds" % (
                         self.colorize('number', '%.3f' % n_seconds, normal))
 
+    def format_seconds_short(self, n_seconds):
+        """Format a time in seconds (short version)."""
+        if n_seconds >= self.slow_test_threshold:
+            color = 'slow-test'
+        else:
+            color = 'number'
+        return self.colorize(color, "%.3f s" % n_seconds)
+
     def summary(self, n_tests, n_failures, n_errors, n_seconds):
         """Summarize the results."""
         sys.stdout.writelines([
@@ -1219,8 +1248,22 @@ def run_layer(options, layer_name, layer, tests, setup_layers,
     if options.resume_layer != None:
         output.info_suboptimal( "  Running in a subprocess.")
 
-    setup_layer(options, layer, setup_layers)
-    return run_tests(options, tests, layer_name, failures, errors)
+    try:
+        setup_layer(options, layer, setup_layers)
+    except EndRun:
+        raise
+    except Exception:
+        f = cStringIO.StringIO()
+        traceback.print_exc(file=f)
+        output.error(f.getvalue())
+        errors.append((SetUpLayerFailure(), sys.exc_info()))
+        return 0
+    else:
+        return run_tests(options, tests, layer_name, failures, errors)
+
+class SetUpLayerFailure(unittest.TestCase):
+    def runTest(self):
+        "Layer set up failure."
 
 def resume_tests(options, layer_name, layers, failures, errors):
     output = options.output
@@ -1314,6 +1357,12 @@ def tear_down_unneeded(options, needed, setup_layers, optional=False):
             output.stop_tear_down(time.time() - t)
         del setup_layers[l]
 
+
+cant_pm_in_subprocess_message = """
+Can't post-mortem debug when running a layer as a subprocess!
+Try running layer %r by itself.
+"""
+
 def setup_layer(options, layer, setup_layers):
     assert layer is not object
     output = options.output
@@ -1324,7 +1373,20 @@ def setup_layer(options, layer, setup_layers):
         output.start_set_up(name_from_layer(layer))
         t = time.time()
         if hasattr(layer, 'setUp'):
-            layer.setUp()
+            try:
+                layer.setUp()
+            except Exception:
+                if options.post_mortem:
+                    if options.resume_layer:
+                        options.output.error_with_banner(
+                            cant_pm_in_subprocess_message
+                            % options.resume_layer)
+                        raise
+                    else:
+                        post_mortem(sys.exc_info())
+                else:
+                    raise
+                    
         output.stop_set_up(time.time() - t)
         setup_layers[layer] = 1
 
@@ -2023,6 +2085,22 @@ reporting.add_option(
 Output progress status
 """)
 
+reporting.add_option(
+    '--no-progress',action="store_false", dest='progress',
+    help="""\
+Do not output progress status.  This is the default, but can be used to
+counter a previous use of --progress or -p.
+""")
+
+# We use a noop callback because the actual processing will be done in the
+# get_options function, but we want optparse to generate appropriate help info
+# for us, so we add an option anyway.
+reporting.add_option(
+    '--auto-progress', action="callback", callback=lambda *args: None,
+    help="""\
+Output progress status, but only when stdout is a terminal.
+""")
+
 reporting.add_option(
     '--color', '-c', action="store_true", dest='color',
     help="""\
@@ -2045,6 +2123,14 @@ reporting.add_option(
 Colorize the output, but only when stdout is a terminal.
 """)
 
+reporting.add_option(
+    '--slow-test', type='float', dest='slow_test_threshold',
+    metavar='N', default=10,
+    help="""\
+With -c and -vvv, highlight tests that take longer than N seconds (default:
+%default).
+""")
+
 reporting.add_option(
     '-1', '--hide-secondary-failures',
     action="store_true", dest='report_only_first_failure',
@@ -2313,14 +2399,23 @@ default_setup_args = [
     '--suite-name', 'test_suite',
     ]
 
+
+def terminal_has_colors():
+    """Determine whether the terminal supports colors.
+
+    Some terminals (e.g. the emacs built-in one) don't.
+    """
+    return tigetnum('colors', -1) >= 8
+
+
 def get_options(args=None, defaults=None):
     # Because we want to inspect stdout and decide to colorize or not, we
     # replace the --auto-color option with the appropriate --color or
     # --no-color option.  That way the subprocess doesn't have to decide (which
-    # it would do incorrectly anyway because stdout wouled be a pipe).
+    # it would do incorrectly anyway because stdout would be a pipe).
     def apply_auto_color(args):
         if args and '--auto-color' in args:
-            if sys.stdout.isatty():
+            if sys.stdout.isatty() and terminal_has_colors():
                 colorization = '--color'
             else:
                 colorization = '--no-color'
@@ -2328,8 +2423,21 @@ def get_options(args=None, defaults=None):
             args[:] = [arg.replace('--auto-color', colorization)
                        for arg in args]
 
+    # The comment of apply_auto_color applies here as well
+    def apply_auto_progress(args):
+        if args and '--auto-progress' in args:
+            if sys.stdout.isatty():
+                progress = '--progress'
+            else:
+                progress = '--no-progress'
+
+            args[:] = [arg.replace('--auto-progress', progress)
+                       for arg in args]
+
     apply_auto_color(args)
     apply_auto_color(defaults)
+    apply_auto_progress(args)
+    apply_auto_progress(defaults)
 
     default_setup, _ = parser.parse_args(default_setup_args)
     assert not _
@@ -2352,6 +2460,7 @@ def get_options(args=None, defaults=None):
 
     if options.color:
         options.output = ColorfulOutputFormatter(options)
+        options.output.slow_test_threshold = options.slow_test_threshold
     else:
         options.output = OutputFormatter(options)
 
@@ -2565,6 +2674,7 @@ def test_suite():
         doctest.DocFileSuite(
         'testrunner-arguments.txt',
         'testrunner-coverage.txt',
+        'testrunner-debugging-layer-setup.test',
         'testrunner-debugging.txt',
         'testrunner-edge-cases.txt',
         'testrunner-errors.txt',
